@@ -906,6 +906,9 @@ ssize_t meta_avtomsg(META *m, META_AV *avlst,
 		/* Binary interface. Skip first 8 bytes and insert hdr later */
 		for(av = avlst, o = buf + 8; av; av = av->next) {
 
+			if ((av->flags & AV_DEL) && 
+			    (flags & AVMSG_ONESHOT) == 0) continue;
+
 			/* Check if av occurs in ACL, if any */
 			if (acl) {
 				for(aclav = acl; 
@@ -968,6 +971,9 @@ ssize_t meta_avtomsg(META *m, META_AV *avlst,
 	/* Ascii interface */
 	for(av = avlst, o = buf; av; av = av->next) {
 
+		if ((av->flags & AV_DEL) && 
+		    (flags & AVMSG_ONESHOT) == 0) continue;
+
 		/* Check if av occurs in ACL, if any */
 		if (acl) {
 			for(aclav = acl; 
@@ -1258,6 +1264,7 @@ void meta_printavlist(META *m, META_AV *avlst, int depth)
 	META_AV *av;
 
 	for(av = avlst; av; av = av->next) {
+		if (av->flags & AV_DEL) continue;
 		meta_printav(m, av, depth);
 		if (av->sub) meta_printavlist(m, av->sub, depth + 1);
 	}
