@@ -44,6 +44,7 @@ char langvm[] = "LANGVM - Copyright (C) 2001 Emile van Bergen.";
 #include <misc.h>
 
 #include <language.h>
+#include <debug.h>
 
 
 /*
@@ -67,6 +68,7 @@ VM *vm_new(INSN *code, ssize_t len, int stackend,
 
 	ret->stk = (META_AV *)malloc(stackend * sizeof(META_AV));
 	if (!ret->stk) { free(ret); return 0; }
+	memset(ret->stk, 0xee, stackend * sizeof(META_AV));
 	ret->head[0] = reqhead; ret->head[1] = rephead;
 	ret->tail[0] = reqtail; ret->tail[1] = reptail;
 	ret->stackend = stackend;
@@ -641,7 +643,7 @@ int vm_run(VM *vm)
 			R.i = 0; R.p = 0; R.l = 1;
 			break;
 
-		  case OP_DELALLAV:
+		  case OP_DELALLAV: if (sp < 1) { ret = VM_UNDERFLOW; break; } 
 		  	/* Fixup USEREP flag here, default to USEREP set */
 			if (!(R.flags & AV_USEREPVALID)) R.flags ^= AV_USEREP;
 
@@ -673,7 +675,8 @@ int vm_run(VM *vm)
 			R.i = 0;
 			break;
 
-		  case OP_MOVEALLAV:
+		  case OP_MOVEALLAV: if (sp < 1) { ret = VM_UNDERFLOW; break; } 
+
 			/* First free right operand's data if any */
 			meta_freeavdata(&R); R.p = 0; R.l = 0;
 
@@ -741,18 +744,37 @@ struct iface *vm_getiface(VM *vm)
 }
 
 
+void vm_dumptrace(VM *vm, META *m, INSN *codestart)
+{
+	msg(F_LANG, L_ERR, "vm: ip=%d limit=%d\n",
+	    vm->ip - codestart, vm->codeend - codestart);
+	msg(F_LANG, L_ERR, "vm: sp=%d limit=%d\n",
+	    vm->sp, vm->stackend);
+	msg(F_LANG, L_ERR, "vm: stack contents:\n");
+	vm_dumpstack(vm, m);
+
+	msg(F_LANG, L_ERR, "vm: request list:\n");
+	meta_printavlist(m, vm->head[VM_REQ], 0);
+	msg(F_LANG, L_ERR, "vm: reply list:\n");
+	meta_printavlist(m, vm->head[VM_REP], 0);
+}
+
+
 void vm_dumpstack(VM *vm, META *m)
 {
-	char buf[8];
+	char buf[16];
 	int n, l;
 
 	if (vm->sp > 0) {
-		write(2, "(top)", 5);
+		write(2, "(top)", 4);
 		meta_printav(m, vm->stk + vm->sp - 1, 0);
 
 		for(n = vm->sp - 2; n >= 0; n--) {
-			buf[0] = '('; l = meta_ordtoa(buf + 1, 3, 2, 10, n);
-			buf[l] = ')'; write(2, buf, l + 1);
+			l = 0; 
+			buf[l++] = '('; 
+			l += meta_ordtoa(buf + l, 3, 2, 10, n);
+			buf[l++] = ')'; 
+			write(2, buf, l);
 			meta_printav(m, vm->stk + n, 0);
 		}
 	}
