@@ -46,6 +46,8 @@
  * 		      to use different installations without editing much.
  * 2002/06/26 - EvB - Fixed trivial compiler warning
  * 2002/10/02 - EvB - Added missing sys/time.h
+ * 2003/06/06 - EvB - Added -2 flag to prevent binding as version 3, which is
+ *		      now the default if the client library supports it
  */
 
 char radldap_id[] = "RADLDAP - Copyright (C) 2001 Emile van Bergen.";
@@ -77,7 +79,7 @@ char radldap_id[] = "RADLDAP - Copyright (C) 2001 Emile van Bergen.";
 #include <platform.h>	/* For U_INT32_T */
 
 
-#define RADLDAP_VERSION	"v0.8b"
+#define RADLDAP_VERSION	"v0.9"
 
 
 #ifndef RADLDAP_MAPFILE
@@ -114,7 +116,7 @@ typedef struct atrmap {
 /* Command line options */
 
 static char *mapfile, *binddn, *password, *host;
-static int port, unbind_after_req, debug;
+static int port, unbind_after_req, no_ldapv3, debug;
 
 
 /*
@@ -145,10 +147,10 @@ void parseoptions(int argc, char **argv)
 	/* Defaults */
 	host = 0; mapfile = RADLDAP_MAPFILE; 
 	binddn = 0; password = "";
-	port = LDAP_PORT; unbind_after_req = debug = 0;
+	port = LDAP_PORT; unbind_after_req = no_ldapv3 = debug = 0;
 
 	/* Handle options */
-	while((c = getopt(argc, argv, "m:b:s:p:udhv")) != -1) {
+	while((c = getopt(argc, argv, "m:b:s:p:u2dhv")) != -1) {
 		switch(c) {
 		  case 'm': mapfile = optarg; break;
 		  case 'b': binddn = optarg; break;
@@ -161,6 +163,7 @@ void parseoptions(int argc, char **argv)
 			}
 			break;
 		  case 'u': unbind_after_req++; break;
+		  case '2': no_ldapv3++; break;
 		  case 'd': debug++; break;
 		  case 'v':
 			fprintf(stderr, "\nRadLDAP module " RADLDAP_VERSION ". "
@@ -286,6 +289,7 @@ int do_bind(LDAP *ld, char *dn, char *pass)
 LDAP *setup_ldap()
 {
 	LDAP *ret;
+	int n;
 
 	if (debug) fprintf(stderr, "radldap: Setting up LDAP for %s, port %d\n",
 			   host, port);
@@ -294,6 +298,15 @@ LDAP *setup_ldap()
 		_exit(2); 
 	}
 
+	/* Set protocol version to 3 if the client library supports it and
+	   the user hasn't asked not to do it */
+#if defined(LDAP_OPT_PROTOCOL_VERSION) && defined(LDAP_VERSION3)
+	if (!no_ldapv3) {
+		n = LDAP_VERSION3;
+		ldap_set_option(ret, LDAP_OPT_PROTOCOL_VERSION, &n);
+	}
+#endif
+
 	/* Bind if we were requested to do so */
 	if (binddn && do_bind(ret, binddn, password)) {
 		ldap_perror(ret, "radldap: Could not bind");
