@@ -43,6 +43,8 @@ char config_id[] = "CONFIG - Copyright (C) 2001 Emile van Bergen.";
 #include <config.h>
 #include <debug.h>
 #include <defs.h>
+#include <jobs.h>	/* for job_del */
+#include <channels.h>	/* for chan_del_jobs */
 
 
 /*
@@ -121,6 +123,7 @@ static int add_iface(CONF *c, META *m, META_AV *data,
 	}
 	memset(iface, 0, sizeof(IFACE));
 	iface->xfertimeout = PRT_XFERTIMEOUT;
+	iface->window = 1;			/* sync iface is default */
 
 	/* And add the interface to the configuration's list */
 	iface->next = c->ifaces;
@@ -139,29 +142,31 @@ static int add_iface(CONF *c, META *m, META_AV *data,
 		  case C_DI_FLAGS: 
 		  	iface->flags = av->l; 
 			break;
-		  
+
 		  case C_DI_TIMEOUT:
 		  	iface->xfertimeout = av->l; 
 			break;
 
+		  case C_DI_WINDOW:
+		  	iface->window = av->l; 
+			break;
+
 		  case C_DI_SENDATTR:
 		  case C_DI_RECVATTR:
+		  case C_DI_JOBTICKET:
 			setspec_n(spec, av->p, av->l);
 			i = meta_getitembyspec(m, spec);
-			if (!i) {
-				msg(F_MISC, L_ERR, "add_iface: ERROR: Unknown "
-						   "ACL attribute '%s'!\n", 
-				    spec, iface->name);
-				return -1;
+			if (!i) { msg(F_MISC, L_ERR, "add_iface: ERROR: Unknown ACL/jobticket attribute '%s'!\n", spec); return -1; }
+
+			if (av->i->nr == C_DI_JOBTICKET) {
+				if (i->val_type != MT_INTEGER && i->val_type != MT_STRING) { msg(F_MISC, L_ERR, "add_iface: ERROR: Job ticket attribute '%s' must be of string or integer type!\n", spec); return -1; }
+				iface->jobticket = i;
+				break;
 			}
 
 			/* Allocate AV item */
 			aclav = (META_AV *)malloc(sizeof(META_AV));
-			if (!aclav) {
-				msg(F_MISC, L_ERR, "add_iface: ERROR: Could "
-						   "not allocate ACL AV!\n");
-				return -1;
-			}
+			if (!aclav) { msg(F_MISC, L_ERR, "add_iface: ERROR: Could not allocate ACL AV!\n"); return -1; }
 			memset(aclav, 0, sizeof(META_AV));
 			aclav->i = i;
 
@@ -178,6 +183,17 @@ static int add_iface(CONF *c, META *m, META_AV *data,
 		}
 	}
 
+	/* If we have a send acl and a job ticket, add the job ticket to
+	   the send acl. A little dwimmy but I think it's worth it here. */
+	if (iface->sendacl && iface->jobticket) {
+		aclav = (META_AV *)malloc(sizeof(META_AV));
+		if (!aclav) { msg(F_MISC, L_ERR, "add_iface: ERROR: Could not allocate ACL AV!\n"); return -1; }
+		memset(aclav, 0, sizeof(META_AV));
+		aclav->i = iface->jobticket;
+		aclav->next = iface->sendacl;
+		iface->sendacl = aclav;
+	}
+
 	/* Check if we at least have a name now */
 	if (!iface->name[0]) {
 		msg(F_MISC, L_ERR, "add_iface: ERROR: 'name' missing from "
@@ -393,7 +409,9 @@ void conf_del(CONF *c)
 	SOCK *s;
 	IFACE *i;
 	CHAN *ch, *tmpch;
+#ifndef CONF_TEST
 	JOB *j, *tmpj;
+#endif
 	PROC *p;
 
 	if (c) {
@@ -405,34 +423,21 @@ void conf_del(CONF *c)
 
 		/* Stop and free interfaces */
 		while(c->ifaces) {
-			
+#ifndef CONF_TEST
 			/* Kill jobs on this interface's shared send queue */
 			for(j = c->ifaces->sendqh; j; j = tmpj) {
-				
-				/* Free encoded message, if any */
-				if (j->pkt) free(j->pkt);
-				if (j->msg) free(j->msg);
-
-				/* Free VM */
-				vm_del(j->vm);
-
-				/* Free job and go to next */
 				tmpj = j->next;
-				free(j);
+				job_del(j);
 			}
-
+			c->ifaces->sendqh = c->ifaces->sendqt = 0;
+#endif
 			/* Free channels */
 			for(ch = c->ifaces->chans; ch; ch = tmpch) {
-
-				/* Kill jobs on this channel's recv. queue */
-				for(j = ch->recvqh; j; j = tmpj) {
-					if (j->pkt) free(j->pkt);
-					if (j->msg) free(j->msg);
-					vm_del(j->vm);
-					tmpj = j->next;
-					free(j);
-				}
-
+#ifndef CONF_TEST
+				/* Kill jobs in this channel's recv. ring 
+				   and free A/V pairs on receive list */
+				chan_flush(ch);
+#endif
 				/* Free channel and go to next */
 				tmpch = ch->next;
 				free(ch);
