@@ -124,7 +124,6 @@ void job_del(JOB *j)
 {
 	if (j) {
 		if (j->pkt) free(j->pkt);
-		if (j->msg) free(j->msg);
 		if (j->vm) {
 			if (j->vm->head[VM_REQ]) 
 				meta_freeavlist(j->vm->head[VM_REQ]);
@@ -206,6 +205,7 @@ IFACE *job_run(CONF *c, JOB *j)
 	  default:			/* The VM choked on the code */
 
 		msg(F_LANG, L_ERR, "job_run: Invalid operation, error code %d - dropping request!\n", n);
+		vm_dumptrace(j->vm, c->m, c->expr);
 		break;
 	}
 	
@@ -261,50 +261,124 @@ void job_toiface(IFACE *i, JOB *j, time_t t)
 
 void job_tochan(CHAN *ch, JOB *j, time_t t)
 {
-	static char reqmsg[C_MAX_MSGSIZE];
+	static char reqmsg[C_MAX_MSGSIZE], jobticketbuf[16];
+	META_AV av;
 	ssize_t msglen;
 	PROC *p;
 
-	/* Put the channel's process tx'er in motion; apparently it was idle */
+	/* For convenience */
 	p = ch->proc;
-	p->state |= PRS_SENDING;
-	p->timer = t + p->xfertimeout;
+
+	/* See if this channel has indeed room (it should, 
+	   otherwise I shouldn't have been called) */
+	if (job_ring_maxput(ch) < 1) {
+		msg(F_PROC, L_ERR, "job_tochan: ERROR: No room in channel's job ring! Dropping request.\n");
+		job_del(j);
+		return;
+	}
+
+	/* Add A/V pair for ch->job_w if we use job tickets */
+	if (ch->iface->jobticket) {
+		memset(&av, 0, sizeof(av));
+		av.i = ch->iface->jobticket;
+		av.l = ch->job_w;
+		if (av.i->val_type == MT_STRING) {
+			av.l = meta_ordtoa(jobticketbuf, sizeof(jobticketbuf), 
+					   0, 10, ch->job_w);
+			av.p = jobticketbuf;
+		}
+		av.prev = j->vm->tail[VM_REQ];
+		if (j->vm->tail[VM_REQ]) j->vm->tail[VM_REQ]->next = &av;
+		else j->vm->head[VM_REQ] = &av;
+	}
 
 	/* Create the message from the request list. */
 	msglen = meta_avtomsg(ch->iface->c->m, j->vm->head[VM_REQ],
 			      reqmsg, C_MAX_MSGSIZE, 
 			      ch->iface->flags, ch->iface->sendacl);
 
+	/* Remove temporary job ticket pair from list tail */
+	if (ch->iface->jobticket) {
+		if (av.prev) av.prev->next = 0;
+		else j->vm->head[VM_REQ] = 0;
+		j->vm->tail[VM_REQ] = av.prev;
+	}
+
 	/* Put the message in the channel's proc's ring. I'd agree we copy 
 	   things around a lot - avtomsg could be done straight to the ring.
 	   The only times meta_avtomsg is used right now is here and in
 	   printav(), which would also benefit from that... */
 	ring_put(p->w, reqmsg, msglen);
 
-	msg(F_PROC, L_DEBUG, "- set sending state on %s's pid %d, %d now in "
-			     "ring\n", 
-	    ch->iface->name, p->pid, ring_maxget(p->w));
+	/* Set expiry time in job */
+	j->expiry = t + ch->iface->xfertimeout;
 
-	/* Put job at head of channel's recv queue */
-	j->next = 0;				/* my next = none */
-	j->prev = ch->recvqh;			/* my prev = current head */
-	if (ch->recvqh) ch->recvqh->next = j;	/* current head's next = me */
-	else ch->recvqt = j;			/* if none, then I'm tail too */
-	ch->recvqh = j;				/* I am always the new head */
+	/* Put the channel's process tx'er in motion; apparently it was idle */
+	p->state |= PRS_SENDING;
+	if (p->timer == 0) p->timer = j->expiry;
+
+	msg(F_PROC, L_DEBUG, "job_tochan: set sending state on %s's pid %d for job %d; %d bytes now in ring; proc timer expires in %d seconds\n", ch->iface->name, p->pid, ch->job_w, ring_maxget(p->w), p->timer - t);
+
+	/* Add job to ring -- we verified that we had room earlier */
+	ch->jobring[ch->job_w] = j;
+	ch->job_w = (ch->job_w + 1) % C_CHAN_JOBRINGSIZE;
 }
 
 
-JOB *job_fromchan(CHAN *ch)
+/* Note: this also takes the received AV list from the channel and
+   adds the pairs on it to the job that will be returned */
+
+JOB *job_fromchan(CHAN *ch, time_t t)
 {
 	JOB *j;
+	META_AV *av;
+	ssize_t len;
+	int nr;
 
-	/* Take job from tail of channel's recv queue */
-	j = ch->recvqt;
-	if (!j) return j;			/* no job: shouldn't happen! */	
+	if (ch->iface->window > 1 && ch->iface->jobticket) {
 
-	ch->recvqt = ch->recvqt->next;		/* advance tail */
-	if (ch->recvqt) ch->recvqt->prev = 0;	/* new tail's prev = none */
-	else ch->recvqh = 0;			/* if no tail, no head either */
+		/* Async iface: scan channel's receive list for a job ticket 
+		   attribute, which must be present, and take its value. */
+
+		for(av = ch->reph; 
+		    av && av->i != ch->iface->jobticket;
+		    av = av->next);
+		if (!av) { msg(F_PROC, L_ERR, "WARNING: Ignoring message received on asynchronous interface: no job ticket present\n"); meta_freeavlist(ch->reph); ch->reph = ch->rept = 0; return 0; }
+
+		nr = av->l;
+		if (av->i->val_type == MT_STRING) {
+			len = 0;
+			nr = meta_atoord(av->p, av->l, 0, 0, &len, 10);
+			if (len != av->l) { msg(F_PROC, L_ERR, "WARNING: Ignoring message containing malformed job ticket '%s'!\n", dbg_cvtstr(av->p, av->l)); meta_freeavlist(ch->reph); ch->reph = ch->rept = 0; return 0; }
+		}
+		msg(F_PROC, L_DEBUG, "- got job ticket %s = %d\n",
+		    av->i->name, nr);
+
+		if (nr < 0 || nr > C_CHAN_JOBRINGSIZE || (j = ch->jobring[nr]) == 0) { msg(F_PROC, L_ERR, "WARNING: Ignoring message for unknown job %d\n", nr); meta_freeavlist(ch->reph); ch->reph = ch->rept = 0; return 0; }
+
+		/* We now have the job in j and its slot number nr in nr. 
+		   Clear the slot in the ring and see how far we can advance 
+		   the read pointer. */
+
+		ch->jobring[nr] = 0;
+		while(ch->jobring[ch->job_r] == 0 && ch->job_r != ch->job_w)
+			ch->job_r = (ch->job_r + 1) % C_CHAN_JOBRINGSIZE;
+	}
+	else {	/* Synchronous interface: just take next job from ring */
+
+		if (ch->job_r == ch->job_w || (j = ch->jobring[ch->job_r]) == 0) { msg(F_PROC, L_ERR, "job_fromchan: ERROR: Ignoring unexpected message from subprocess; no job in queue!\n"); meta_freeavlist(ch->reph); ch->reph = ch->rept = 0; return 0; }
+
+		ch->jobring[ch->job_r] = 0;
+		ch->job_r = (ch->job_r + 1) % C_CHAN_JOBRINGSIZE;
+	}
+
+	/* Move received list of A/V pairs to job's reply list */
+
+	if (ch->reph) ch->reph->prev = j->vm->tail[VM_REP];
+	if (j->vm->tail[VM_REP]) j->vm->tail[VM_REP]->next = ch->reph;
+	else j->vm->head[VM_REP] = ch->reph;
+	j->vm->tail[VM_REP] = ch->rept;
+	ch->reph = ch->rept = 0;
 
 	return j;
 }
