@@ -11,6 +11,8 @@
  *
  * History:
  * 2001/06/26 - EvB - Created
+ * 2004/07/01 - EvB - Adapted to allow channels to have a window size > 1
+ *                    making asynchronous modules possible
  */
 
 
@@ -28,7 +30,7 @@
 #include <netinet/in.h>	/* For sockaddr_in */
 
 #include <subprocs.h>	/* For PROC */
-#include <metadata.h>	/* For META_ORD, META_NAME */
+#include <metaops.h>	/* For META_ORD, META_NAME (metadata), META_AV */
 
 
 /*
@@ -47,19 +49,30 @@ typedef struct sock {
 typedef struct job {
 	struct job *next, *prev;
 	struct vm *vm;
-	char *pkt, *msg;	/* msg is used while on shared send queue, to */
-	ssize_t pktlen, msglen;	/* prevent re-encoding when finding a channel */
-	int replyfd;
-	struct sockaddr_in replyadr;
+	char *pkt;
+	ssize_t pktlen;
+	int replyfd;		/* socket we'll reply from */
+	struct sockaddr_in replyadr;	/* address we'll reply to */
+	time_t expiry;		/* expiry time if assigned to a channel. Used
+				   to guard module subprocesses. */
 } JOB;
 
+#define job_ring_empty(ch)	((ch)->job_r == (ch)->job_w)
+#define job_ring_maxput(ch)	((C_CHAN_JOBRINGSIZE + (ch)->job_r - ((ch)->job_w + 1)) % C_CHAN_JOBRINGSIZE)
+#define job_ring_maxget(ch)	((C_CHAN_JOBRINGSIZE + (ch)->job_w - (ch)->job_r) % C_CHAN_JOBRINGSIZE)
 
 typedef struct chan {
 	struct chan *next;
 	struct iface *iface;	/* inited */
-	JOB *recvqh, *recvqt;
+	META_AV *reph, *rept;	/* received reply list */
+	int job_r, job_w;	/* read and write indexes into job ring */
+	JOB *jobring[C_CHAN_JOBRINGSIZE];
 	PROC *proc;		/* inited - needed to give the right proc a */
-				/* kick if it's idle and a new req. appears */
+				/* kick if it's idle and a new req. appears;
+				   also to reset the watchdog timer, which
+				   now supervises the advancing of the job
+				   ring read pointer instead of reception of 
+				   single messages. */
 } CHAN;
 
 
@@ -68,9 +81,12 @@ typedef struct iface {
 	struct conf *c;		/* inited */
 	META_NAME name;		/* inited */
 	JOB *sendqh, *sendqt;
+	int sendqlen;
 	struct meta_av *sendacl, *recvacl;	/* inited */
 	CHAN *chans, *rrch;	/* inited */
 	int xfertimeout, flags;	/* inited - only really used to pass to procs */
+	int window;		/* inited */
+	META_ITEM *jobticket;	/* inited */
 } IFACE;
 
 
