@@ -2,8 +2,8 @@
  * RADIUS CLIENT - Sends authentication or accounting requests to RADIUS
  *		   servers and displays the results. I/O is done according to
  * 		   OpenRADIUS' module interface, so this can also be used for
- *		   trivial proxying. Operates fully asynchronously; handles
- *		   multiple queries and retransmissions for each in parallel.
+ *		   proxying. Operates fully asynchronously; handles multiple 
+ *		   queries and retransmissions for each in parallel.
  *
  * Emile van Bergen, emile@evbergen.xs4all.nl
  *
@@ -16,6 +16,10 @@
  * 2002/10/18 - EvB - Started, based on the old radclient test tool, Brian
  *                    Candler's enhancements and my RADIUS client for pppd.
  * 2003/04/01 - EvB - Fixed typo in reply matching code
+ * 2004/06/26 - EvB - Added Target-Server attribute support, i.e. ability
+ *                    to specify servers in query instead of command line
+ * 2004/08/07 - EvB - Split -C from -p option; you do want to reencode PAP
+ *                    but don't want to reencode CHAP when proxying.
  */
 
 
@@ -93,28 +97,29 @@
 #define A_CHAP_PASS		4		/* query */
 #define A_CHAP_CHAL		8		/* query */
 #define A_NAS_IP		0x10		/* query */
+#define A_TGT_SRV		0x20		/* query */
 
-#define A_QRY_MASK		0x1f		/* pairs to be added to qry */
+#define A_QRY_MASK		0x3f		/* pairs to be added to qry */
 
-#define A_RAD_ID		0x20		/* request */
-#define A_RAD_AUTH		0x40		/* request */
-#define A_USER_PASS		0x80		/* request (but have given_~) */
-#define A_ACCT_DLY		0x100		/* request */
+#define A_RAD_ID		0x40		/* request */
+#define A_RAD_AUTH		0x80		/* request */
+#define A_USER_PASS		0x100		/* request (but have plain) */
+#define A_ACCT_DLY		0x200		/* request */
 
-#define A_REQ_MASK		0x1e0		/* pairs to be added to req */
+#define A_REQ_MASK		0x3c0		/* pairs to be added to req */
 
-#define A_ALL			0x1ff		/* all pairs that we can add */
+#define A_ALL			0x3ff		/* all pairs that we can add */
 
 /* pairs missing from response */
 
-#define A_RAD_LEN		0x200		/* response */
+#define A_RAD_LEN		0x400		/* response */
 
 
 /* utility macros */
 
 #define tmr_ring_maxput(c) ((MAX_OUTREQS + (c)->timer_r - \
 					  ((c)->timer_w + 1)) % MAX_OUTREQS)
-#define tmr_ring_maxget(c) ((MAX_OUTREQS + (c)->timer_r - (c)->timer_w) % \
+#define tmr_ring_maxget(c) ((MAX_OUTREQS + (c)->timer_w - (c)->timer_r) % \
 			    MAX_OUTREQS)
 
 /*
@@ -126,6 +131,7 @@ struct radsrv {			/* RADIUS server */
 
 	struct sockaddr_in sin;		/* IP address, port */
 	unsigned char *secret;		/* shared secret */
+	unsigned int secretlen;
 
 	int req_slot;			/* -1 or last request to this server */
 };
@@ -177,7 +183,7 @@ struct radcfg {
 	META_SPC *ds_rad_pkt;
 	META_ITEM *di_int, *di_rad_code, *di_rad_id, *di_rad_len, *di_rad_auth;
 	META_ITEM *di_user_name, *di_user_pass, *di_chap_pass, *di_chap_chal; 
-	META_ITEM *di_nas_ip, *di_acct_dly, *di_query_id;
+	META_ITEM *di_nas_ip, *di_acct_dly, *di_query_id, *di_tgt_srv;
 	META_VAL *dv_auth_req, *dv_acct_req;
 	META_ORD nas_ip_address;
 	int max_idbits, max_ids, max_idmask;
@@ -187,7 +193,7 @@ struct radcfg {
 	char *raddb;
 	struct sockaddr_in sin_src;
 	int retrycount, timeout, doacct;
-	int nostatus, ioflags, noencpass; 
+	int nostatus, ioflags, noencpap, noencchap; 
 	int flood, outimm, quiet, debug;
 	struct radsrv srv[MAX_SERVERS];	/* list of configured servers */
 	int srvcnt;
@@ -266,9 +272,10 @@ void query_del(struct radcfg *c, struct radquery *qry)
 		for(req_slot = qry->srv[srvnr].req_slot;
 		    req_slot != -1;
 		    req_slot = c->req[req_slot].prev_slot) {
+			if (c->req[req_slot].pktlen == 0) continue;
+			c->req[req_slot].pktlen = 0;
 			msg(F_RECV, L_NOTICE, "Done with slot %d, id %d\n",
 			    req_slot, c->req[req_slot].id);
-			c->req[req_slot].pktlen = 0;
 		}
 	}
 
@@ -419,7 +426,7 @@ void sock_handle_read(struct radcfg *c, time_t t)
 	md5_append(&md5ctx, r->auth, sizeof(r->auth));
 	md5_append(&md5ctx, rad_auth->p + rad_auth->l,
 			    rad_len - (rad_auth->p + rad_auth->l - pkt));
-	md5_append(&md5ctx, r->srv->secret, strlen(r->srv->secret));
+	md5_append(&md5ctx, r->srv->secret, r->srv->secretlen);
 	md5_finish(&md5ctx, chkauth);
 
 	if (rad_auth->l != sizeof(chkauth) ||
@@ -531,7 +538,7 @@ int new_req(struct radcfg *c, time_t t, struct radquery *qry,
 		av->p = safe_malloc((qry->given_pap_pass->l + 15) & ~15);
 		encrypt_attr_pap(qry->given_pap_pass->p, qry->given_pap_pass->l,
 				 av->p, &n, 
-				 srv->secret, strlen(srv->secret),
+				 srv->secret, srv->secretlen,
 				 r->auth, sizeof(r->auth));
 		av->l = n;
 		av->flags |= AV_FREE_P;
@@ -590,7 +597,7 @@ int new_req(struct radcfg *c, time_t t, struct radquery *qry,
 	    } else {
 		md5_init(&md5ctx);
 		md5_append(&md5ctx, r->pkt, r->pktlen);
-		md5_append(&md5ctx, srv->secret, strlen(srv->secret));
+		md5_append(&md5ctx, srv->secret, srv->secretlen);
 		md5_finish(&md5ctx, r->auth);
 	    }
 	    memcpy(r->pkt + c->di_rad_auth->val_ofs, r->auth, 
@@ -656,8 +663,9 @@ void send_qry_req(struct radcfg *c, time_t t, struct radquery *qry)
 		}
 		else if (oldsrv && oldsrv->req_slot != -1 &&
 		         c->req[oldsrv->req_slot].pktlen &&
-			 (c->noencpass || qry->given_pap_pass == 0 ||
-			  strcmp(oldsrv->secret, srv->secret) == 0)) {
+			 (c->noencpap || qry->given_pap_pass == 0 ||
+			  (oldsrv->secretlen == srv->secretlen &&
+			   memcmp(oldsrv->secret, srv->secret, srv->secretlen) == 0))) {
 			slot = oldsrv->req_slot;
 		}
 	}
@@ -711,7 +719,10 @@ int fixup_query(struct radcfg *c, struct radquery *q)
 {
 	META_AV *av, *nextav, *chap_chal, *chap_pass;
 	md5_state_t md5ctx;
-	char chap_id;
+	struct radsrv *srv;
+	char chap_id, *s;
+	int l, n, port;
+	U_INT32_T host;
 
 	/* See which attributes we have, i.e. that don't need to be added.
 	   Also, save some pointers to pairs we want to reference later. 
@@ -719,6 +730,7 @@ int fixup_query(struct radcfg *c, struct radquery *q)
 
 	chap_pass = chap_chal = 0;
 	q->missing = A_ALL;
+	q->acct = c->doacct;
 	for(av = q->reqhead; av; av = nextav) {
 		nextav = av->next;
 		if (av->i == c->di_rad_code)	   q->missing &= ~ A_RAD_CODE,
@@ -744,12 +756,37 @@ int fixup_query(struct radcfg *c, struct radquery *q)
 			meta_remav(&q->reqhead, &q->reqtail, av);
 			meta_addav(&q->rephead, &q->reptail, 0, 0, av);
 		}
+		else if (av->i == c->di_tgt_srv && q->srvcnt < MAX_SERVERS) { 
+			srv = &q->srv[q->srvcnt];
+
+		 	srv->req_slot = -1;
+			srv->sin.sin_family = AF_INET;
+			srv->sin.sin_port = htons(q->acct ? DEF_ACCTPORT 
+							  : DEF_AUTHPORT);
+			if ((s = memchr(av->p, '/', av->l)) == 0 || 
+			    (l = av->p + av->l - s) <= 1) { msg(F_PROC, L_ERR, "Warning: ignoring malformed Target-Server attribute (no secret)\n"); continue; }
+			srv->secret = s + 1; 
+			srv->secretlen = l - 1;
+			av->l = (s - av->p);
+			if ((s = memchr(av->p, ':', av->l))) {
+				l = av->p + av->l - s - 1;
+				port = meta_atoord(s + 1, l, 0, 0, &n, 10);
+				if (l == 0 || l > 5 || n != l || port < 1 || port > 65535) { msg(F_PROC, L_ERR, "Warning: ignoring malformed Target-Server attribute (invalid port)\n"); continue; }
+				srv->sin.sin_port = htons(port);
+				av->l = (s - av->p);
+			}
+			host = meta_atoip(av->p, av->l, 0, 0, &n);
+			if (n != av->l || host == 0 || host == 0xffffffff) { msg(F_PROC, L_ERR, "Warning: ignoring malformed Target-Server attribute (invalid IP address)\n"); continue; }
+			srv->sin.sin_addr.s_addr = htonl(host);
+
+			q->srvcnt++;
+			q->missing &= ~ A_TGT_SRV;
+		}
 	}
 
 	/* Fix up missing RAD-Code (and accounting flag) */
 
 	if (q->missing & A_RAD_CODE) {
-		q->acct = c->doacct;
 		av = safe_malloc(sizeof(META_AV));
 		memset(av, 0, sizeof(META_AV));
 		av->i = c->di_rad_code; 
@@ -792,7 +829,7 @@ int fixup_query(struct radcfg *c, struct radquery *q)
 		av->l = c->nas_ip_address;
 		meta_addav(&q->reqhead, &q->reqtail, 0, 0, av);
 	}
-	if (q->missing & A_CHAP_CHAL && c->noencpass == 0) {
+	if (q->missing & A_CHAP_CHAL && c->noencchap == 0) {
 		av = safe_malloc(sizeof(META_AV));
 		memset(av, 0, sizeof(META_AV));
 		av->i = c->di_chap_chal;
@@ -801,10 +838,15 @@ int fixup_query(struct radcfg *c, struct radquery *q)
 		meta_addav(&q->reqhead, &q->reqtail, 0, 0, av);
 		chap_chal = av;
 	}
+	if (q->missing & A_TGT_SRV) {
+		if (c->srvcnt == 0) { msg(F_PROC, L_ERR, "ERROR: No target servers specified!\n"); return -1; }
+		q->srvcnt = c->srvcnt;	/* Use servers from cmd line */
+		memcpy(q->srv, c->srv, sizeof(q->srv));
+	}
 
 	/* Encode the CHAP password here if so requested */
 
-	if (c->noencpass == 0 && chap_pass && chap_chal) {
+	if (c->noencchap == 0 && chap_pass && chap_chal) {
 		get_random_data(&chap_id, 1);
 		md5_init(&md5ctx);
 		md5_append(&md5ctx, &chap_id, 1);
@@ -820,11 +862,11 @@ int fixup_query(struct radcfg *c, struct radquery *q)
 		md5_finish(&md5ctx, chap_pass->p + 1);
 	}
 
-	/* If we have a User-Password and noencpass == 0, tell new_request that
+	/* If we have a User-Password and noencpap == 0, tell new_request that
 	   we're missing an encoded User-Password and take the cleartext 
 	   version off the list, remembering it only in the query structure. */
 
-	if (c->noencpass == 0 && q->given_pap_pass) {
+	if (c->noencpap == 0 && q->given_pap_pass) {
 		q->missing |= A_USER_PASS;
 		meta_remav(&q->reqhead, &q->reqtail, q->given_pap_pass);
 	}
@@ -849,8 +891,6 @@ struct radquery *add_new_query(struct radcfg *c, time_t t,
 	ret->reqtail = reqtail;
 	ret->timestamp = t;
 	ret->ttl = c->retrycount;
-	ret->srvcnt = c->srvcnt;	/* Copy servers from command line */
-	memcpy(ret->srv, c->srv, sizeof(ret->srv));
 	ret->cursrv = -1;		/* No current server yet */
 
 	/* Fixup query, reply if we cannot */
@@ -942,13 +982,14 @@ void usage(char *msg)
 	if (msg) fprintf(stderr, "\nERROR: %s\n\n", msg);
 	fprintf(stderr, 
 		"Usage: radclient [-r raddb] [-i sourceip] [-c retrycount]\n"
-		"                 [-t timeout] [-anlpofqv]\n"
+		"                 [-t timeout] [-anlpCofqv]\n"
 		"                 server[:port] secret [server[:port] secret]...\n"
 		"Options:\n\n"
 		"  -a	default to Accounting if no RAD-Code given\n"
 		"  -n	suppress result code attribute in output\n"
 		"  -l	use long attribute names in output\n"
-		"  -p	send given User-Password or CHAP-Password as-is\n"
+		"  -p	send given User-Password as-is; must be pre-encoded\n"
+		"  -C	send given CHAP-Password and CHAP-Challenge as-is\n"
 		"  -o	output response immediately, don't retry till valid\n"
 /*		"  -f	flood, send retrycount packets as fast as possible\n" */
 		"  -q	quiet, no output on stderr\n"
@@ -975,7 +1016,7 @@ void parseoptions(int argc, char **argv, struct radcfg *c)
 		     AVMSG_NAMEDCONST | AVMSG_SHORTATTR;
 
 	for(fac = 0; fac < F_CNT; fac++) msg_setthresh(fac, L_NOTICE);
-	while((o = getopt(argc, argv, "r:i:c:t:anlpofqvd:D:h")) != -1) {
+	while((o = getopt(argc, argv, "r:i:c:t:anlpCofqvd:D:h")) != -1) {
 
 		switch(o) {
 		  case 'r': c->raddb = optarg; break;
@@ -996,7 +1037,8 @@ void parseoptions(int argc, char **argv, struct radcfg *c)
 		  case 'a': c->doacct = 1; break;
 		  case 'n': c->nostatus = 1; break;
 		  case 'l': c->ioflags &= ~AVMSG_SHORTATTR; break;
-		  case 'p': c->noencpass = 1; break;
+		  case 'p': c->noencpap = 1; break;
+		  case 'C': c->noencchap = 1; break;
 		  case 'o': c->outimm = 1; break;
 /*		  case 'f': c->flood = 1; break;		*/
 		  case 'q': for(fac = 0; fac < F_CNT; fac++) 
@@ -1012,8 +1054,6 @@ void parseoptions(int argc, char **argv, struct radcfg *c)
 		}
 	}
 
-	if (optind + 1 >= argc) usage("Missing server(s)!");
-
 	while(optind < argc && c->srvcnt < MAX_SERVERS) {
 
 		srv = &c->srv[c->srvcnt++];
@@ -1036,7 +1076,8 @@ void parseoptions(int argc, char **argv, struct radcfg *c)
 		}
 
 		if (++optind >= argc || (srv->secret = argv[optind++]) == 0 || 
-		    srv->secret == 0) usage("Missing secret!");
+		    *srv->secret == 0) usage("Missing secret!");
+		srv->secretlen = strlen(srv->secret);
 	}
 
 	msg_init(-1, c->quiet);
@@ -1077,6 +1118,8 @@ void opendict(struct radcfg *c)
 			    			   "Acct-Delay-Time", 0)) ||
 	    !(c->di_query_id  = meta_getitembyname(c->m, 0, 
 			    			   "Radclient-Query-Id", 0)) ||
+	    !(c->di_tgt_srv   = meta_getitembyname(c->m, 0, 
+			    			   "Target-Server", 0)) ||
 	    !(c->dv_auth_req  = meta_getvalbyname(c->m, c->di_rad_code,
 			    			  "Access-Request")) ||
 	    !(c->dv_acct_req  = meta_getvalbyname(c->m, c->di_rad_code,
